#define PATH_MAX        4096	/* # chars in a path name including nul */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/file.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <list>
#include <string>
#include <vector>



#include "trace_v2.h"


/* The older version of this trace file is trace.cpp in path /ClientEngine/base/ctxwebhelper
This v2 version was made for tracing to be much more modular (which can be used by multiple modules)
This version has it's own getICARoot function instead of depending on the module to pass ICAROOT

Usage:
    You need to first call init() before calling logInfo


*/


bool trace::checkSection(const char *section, char *cp)
{
    char *ep;

    while (*cp == ' ' || *cp == '\t')
        ++cp;
    ep = cp;
    while (*ep != ']' && *ep != '\n' && *ep)
        ++ep;
    if (*ep != ']')
        return false;

    /* Got brackets, trim white. */

    --ep;
    while (ep >= cp && (*ep == ' ' || *ep == '\t'))
        --ep;
    if (ep < cp)
        return false;
    return !strncasecmp(section, cp, ep - cp + 1);
}

bool trace::getLogFlag(const char *file_name, const char *section)
{
    FILE *fp;
    char buf[1024];
    char *cp, *ep;
    bool in_section = false;
    bool got_eq;
    bool ret = false;

    fp = fopen(file_name, "r");
    if (!fp)
        return false;
    if (flock(fileno(fp), LOCK_EX))
    {
        perror("error in file locking");
        fclose(fp);
        return false;
    }
    if (fflush(fp))
    {
        perror("error in flushing file");
        // return false;
    }
    while (fgets(buf, sizeof buf, fp))
    {

        cp = buf;
        while (*cp == ' ' || *cp == '\t')
            ++cp;
        if (*cp == '\0' || *cp == '\n' || *cp == ';' || *cp == '#')
            continue; /* Blank or comment - ignore. */
        ep = cp + 1;
        if (section)
        {
            if (*cp == '[')
            {
                in_section = trace::checkSection(section, ep);
                //This continue is to take the last section. If there are multiple sections then this makes sure that it is the last one.
                continue;
            }
            if (!in_section)
                continue;
        }

        /* Look for key = value. */
        while (*ep && *ep != '=' && *ep != ' ' && *ep != '\t' && *ep != '\n') 
        {
            ++ep;
        }

        if (!*ep || *ep == '\n') /* Just a word */
            continue;
        got_eq = (*ep == '=');

        /* Check for a known parameter name. */
        *ep = '\0';
        if (logSwitchString != std::string(cp))
            continue;       /* Unknown parameter - ignore. */

        /* Clear white-space. */
        cp = ep + 1;
        while (*cp == ' ' || *cp == '\t')
            ++cp;

        if (!got_eq)
        {
            if (*cp != '=')
                continue; /* No equals sign - ignore. */
            do
                ++cp;
            while (*cp == ' ' || *cp == '\t');
        }
        if (*cp == '\n' || !*cp)
            continue; /* No RHS. */

        /* Concatentate lines ending with a backslash. */

        ep = cp;
        for (;;)
        {
            ep = ep + strlen(ep) - 1;
            if (*ep != '\n')
                break;
            *ep-- = '\0';
            if (*ep != '\\')
                break;
            if (ep >= buf + (sizeof buf) - 2 ||
                    !fgets(ep, buf + (sizeof buf) - ep, fp))
            {
                *ep-- = '\0';
                break;
            }
        }

        /* Strip trailing white-space. */

        while (ep > cp && (*ep == ' ' || *ep == '\t'))
            *ep-- = '\0';

        ret = (bool)atoi(cp);

        break;
    }

    if (fflush(fp))
    {
        perror("file flush error");
    }
    if (flock(fileno(fp), LOCK_UN))
    {
        perror("flock error");
    }

    fclose(fp);
    return ret;
}

std::string trace::getICARoot(){
    
    std::string icarootDir;
	const char* icarootEnv = getenv("ICAROOT");
	if(icarootEnv)
	{
		icarootDir = std::string(icarootEnv);
		if(ifValidICAROOT(icarootDir))
			return icarootDir;
	}
	
	std::string icaRootDir = GetModulePath();
    // This includes the executable name, we're only interested in the directory part
    size_t found = icaRootDir.find_last_of("/");
    if (found != std::string::npos) {
        icaRootDir = icaRootDir.substr(0, found);
    }
    if (ifValidICAROOT(icaRootDir)) {
        return icaRootDir;
    } else {
        // As a utility see if this binary is running in the util directory
        // of an ICAROOT directory
        found = icaRootDir.find_last_of("/");
        if (found != std::string::npos) {
            if (icaRootDir.substr(found +1).compare("util") == 0) {
                icaRootDir = icaRootDir.substr(0, found);
                if (ifValidICAROOT(icaRootDir)) {
                    return icaRootDir;
                }
            }
        }
    }
    
    // Check the default installation location
    // TODO should we hardcode this?
    icaRootDir = "/opt/Citrix/ICAClient";
    if (ifValidICAROOT(icaRootDir)) {
        return icaRootDir;
    }
    
    // Check the old default installation location, some OEMs prefer to use this still
    // presumably because they have a lot of hardcoded references to it.
    // TODO should we hardcode this?
    icaRootDir = "/usr/lib/ICAClient";
    if (ifValidICAROOT(icaRootDir)) {
        return icaRootDir;
    }
    
    std::cerr<<"No valid ICAROOT directory found"<<std::endl;
	return "";
    
}

bool trace::ifFileExists(std::string &configFile) {
	
	struct stat fileStatus;
	int iretStat = stat(configFile.c_str(), &fileStatus);
	
	if(iretStat == 0)
		return  true;
	
	return false ;
}

std::string trace::GetModulePath()
{
	std::string path;
	char buf[PATH_MAX];
	ssize_t len = readlink("/proc/self/exe", buf, sizeof(buf)-1);
	if (len != -1) {
		buf[len] = '\0';
		path = buf ;
	}
	else {
		path = "<unknown>";
	}

	return path;
}

bool trace::ifValidICAROOT(std::string dir)
{
	if(!dir.length()) return false;

	std::vector<std::string> expectedFiles(4);
	expectedFiles[0] = std::string("AuthManagerDaemon");
	expectedFiles[1] =  std::string("ServiceRecord");
	expectedFiles[2] = std::string("PrimaryAuthManager");
	expectedFiles[3] = std::string("util/storebrowse");
	std::string file;
	for(std::vector<std::string>::iterator it = expectedFiles.begin() ; it != expectedFiles.end() ; it++)
	{
		file = dir+"/"+*it;
		if(!ifFileExists(file))
			return false;
	}
	return true;
}


trace::trace(){

}

/* 
    This init() function will take the section variable, look for section in debug.ini file and set the logSwitch variable
    If this logSwitch variable is 1 then logInfo will send the logs, else no.
    Make sure your ICAROOT variable is set to get the path of debug.ini 
*/

trace trace::init(char* l_section, char* icaroot, char* logSwitchString){

    trace t1;
    t1.logSwitchString = std::string(logSwitchString);
    t1.section = std::string(l_section);
    t1.logSwitch = false; //Default value
    t1.logSwitchRead = false; //Default value
    std::string conf_file_path;
    //Note that it is recommended for the processes calling init() to pass icaroot 
    //See trace::getICARoot() 
    if(icaroot == NULL){
        conf_file_path = trace::getICARoot();
    }
    else{
        conf_file_path = std::string(icaroot);
    }

    //If logInfo called for the first time then call getLogFlag
    if (!t1.logSwitchRead)
    {
        t1.logSwitchRead = true;
        //Get the ICAROOT env variable
        if(conf_file_path.empty()){
            printf("Could not find ICAROOT variable. Failing. Please set the ICAROOT variable to see the logs\n");
            return t1;
        }

        conf_file_path += "/debug.ini";
        
        // printf("section not provided. Please provide the section by using trace::setSection method\n");
        if(!t1.section.empty()){
            t1.logSwitch = t1.getLogFlag(conf_file_path.c_str(), t1.section.c_str());
        }
        else{
            std::cerr << "Section: " << l_section << "not found in debug.ini\n";
        }
        
    }

    return t1;
}



void trace::logInfo(const char *fmt, ...)
{
    
#ifndef LOG_CHECK_PRINTF // Define this macro when compiling trace_v2.cpp
                        // to get the logs on stdout to verify the logs
    if(!logSwitchRead){
        printf("trace::init() function not called. Call init before calling any trace functions\n");
        return;
    }
    
    //If log_flag off then return
    if (!logSwitch)
    {
        return;
    }

    //If trace::openLogCalled = false then call openLog
    if (!openLogCalled)
    {
        openLogCalled = true;
        openlog(0, LOG_PID, LOG_USER);
    }
#endif
    va_list vargs;
    va_start(vargs, fmt);
    syslog(LOG_INFO, "%s%s", section.c_str(), ": ");
    vsyslog(LOG_INFO, fmt, vargs);
    va_end(vargs);

    return;
};
